linux-fsdevel.vger.kernel.org archive mirror
 help / color / mirror / Atom feed
From: Amir Goldstein <amir73il@gmail.com>
To: Jan Kara <jack@suse.cz>
Cc: Al Viro <viro@zeniv.linux.org.uk>, Eric Paris <eparis@redhat.com>,
	linux-fsdevel@vger.kernel.org, linux-kernel@vger.kernel.org
Subject: [RFC][PATCH 2/2] fsnotify: implement event reporting to super block's root inode
Date: Tue, 20 Dec 2016 17:20:07 +0200	[thread overview]
Message-ID: <1482247207-4424-3-git-send-email-amir73il@gmail.com> (raw)
In-Reply-To: <1482247207-4424-1-git-send-email-amir73il@gmail.com>

When the flag FS_EVENT_ON_SB is set on a super block's root inode
fsnotify mask, all events on inodes that are on the same super
block are reported to the root inode.

Events reported to root inode carry the flag FS_EVENT_ON_SB,
to distinguish from events that happen to the root inode itself
or to its direct children (FS_EVENT_ON_CHILD).

The extra cost for vfs operations without any root inode watch
is the test for fsnotify flag on root inode, i.e.:
(dentry->d_sb->s_root->d_inode->i_fsnotify_mask & FS_EVENT_ON_SB)
This could be further optimized by setting a flag on the super_block
struct.

Signed-off-by: Amir Goldstein <amir73il@gmail.com>
---
 fs/notify/fsnotify.c             | 58 +++++++++++++++++++++++++++++++++++++---
 include/linux/fsnotify_backend.h | 10 +++++++
 2 files changed, 64 insertions(+), 4 deletions(-)

diff --git a/include/linux/fsnotify_backend.h b/include/linux/fsnotify_backend.h
index b7992da..224c4aa 100644
--- a/include/linux/fsnotify_backend.h
+++ b/include/linux/fsnotify_backend.h
@@ -266,6 +266,16 @@ extern void __fsnotify_inode_delete(struct inode *inode);
 extern void __fsnotify_vfsmount_delete(struct vfsmount *mnt);
 extern u32 fsnotify_get_cookie(void);
 
+static inline int fsnotify_inode_watches_sb(struct inode *inode)
+{
+	/* FS_EVENT_ON_SB is set if the sb root inode may care */
+	if (!(inode->i_fsnotify_mask & FS_EVENT_ON_SB))
+		return 0;
+	/* this root inode might care about sb events, does it care about the
+	 * specific set of events that can happen on a distant child? */
+	return inode->i_fsnotify_mask & FS_EVENTS_POSS_ON_SB;
+}
+
 static inline int fsnotify_inode_watches_children(struct inode *inode)
 {
 	/* FS_EVENT_ON_CHILD is set if the inode may care */
diff --git a/fs/notify/fsnotify.c b/fs/notify/fsnotify.c
index 12d4479..c0a596c 100644
--- a/fs/notify/fsnotify.c
+++ b/fs/notify/fsnotify.c
@@ -183,12 +183,14 @@ static int send_to_group(struct inode *to_tell,
 
 /*
  * This is the main call to fsnotify.  The VFS calls into hook specific functions
  * in linux/fsnotify.h.  Those functions then in turn call here.  Here will call
  * out to all of the registered fsnotify_group.  Those groups can then use the
  * notification event in whatever means they feel necessary.
  */
-int fsnotify(struct inode *to_tell, __u32 mask, void *data, int data_is,
-	     const unsigned char *file_name, u32 cookie)
+static int __fsnotify(struct inode *to_tell, __u32 mask,
+		      void *data, int data_is,
+		      const unsigned char *file_name, u32 cookie)
 {
 	struct hlist_node *inode_node = NULL, *vfsmount_node = NULL;
 	struct fsnotify_mark *inode_mark = NULL, *vfsmount_mark = NULL;
@@ -196,7 +198,7 @@ int fsnotify(struct inode *to_tell, __u32 mask, void *data, int data_is,
 	struct mount *mnt;
 	int idx, ret = 0;
 	/* global tests shouldn't care about events on child only the specific event */
-	__u32 test_mask = (mask & ~FS_EVENT_ON_CHILD);
+	__u32 test_mask = (mask & ~FS_EVENT_ON_DESCENDANT);
 
 	if (data_is == FSNOTIFY_EVENT_PATH)
 		mnt = real_mount(((struct path *)data)->mnt);
@@ -291,6 +293,54 @@ int fsnotify(struct inode *to_tell, __u32 mask, void *data, int data_is,
 
 	return ret;
 }
+
+/* Notify this dentry's sb root about descendant inode events. */
+static int fsnotify_root(struct dentry *dentry, __u32 mask,
+			 void *data, int data_is,
+			 const unsigned char *file_name, u32 cookie)
+{
+	struct inode *r_inode = d_inode(dentry->d_sb->s_root);
+
+	if (likely(!fsnotify_inode_watches_sb(r_inode)) ||
+	    !(r_inode->i_fsnotify_mask & mask))
+		return 0;
+
+	/* we are notifying root so come up with the new mask which
+	 * specifies these are events which came from sb. */
+	mask |= FS_EVENT_ON_SB;
+
+	return __fsnotify(r_inode, mask, data, data_is, file_name, cookie);
+}
+
+/* Notify this inode and maybe the sb root inode. */
+int fsnotify(struct inode *to_tell, __u32 mask, void *data, int data_is,
+	     const unsigned char *file_name, u32 cookie)
+{
+	struct dentry *dentry = NULL;
+	int ret = 0;
+
+	BUG_ON(mask & FS_EVENT_ON_SB);
+
+	if (data_is == FSNOTIFY_EVENT_PATH)
+		dentry = ((struct path *)data)->dentry;
+	else if (data_is == FSNOTIFY_EVENT_DENTRY)
+		dentry = data;
+
+	if (dentry) {
+		/* First, notify root inode if it cares */
+		ret = fsnotify_root(dentry, mask, data, data_is,
+				    file_name, cookie);
+		if (ret)
+			return ret;
+
+		/* Do not report to root sb watch an event twice */
+		if (unlikely(fsnotify_inode_watches_sb(to_tell)))
+			return 0;
+	}
+
+	/* Then, notify this inode */
+	return __fsnotify(to_tell, mask, data, data_is, file_name, cookie);
+}
 EXPORT_SYMBOL_GPL(fsnotify);
 
 static __init int fsnotify_init(void)
@@ -183,12 +183,14 @@ static int send_to_group(struct inode *to_tell,
 /*
  * This is the main call to fsnotify.  The VFS calls into hook specific functions
- * in linux/fsnotify.h.  Those functions then in turn call here.  Here will call
+ * in linux/fsnotify.h.  Those functions call the helpers fsnotify(),
+ * fsnotify_root(), fsnotify_parent() and they in turn call here. Here will call
  * out to all of the registered fsnotify_group.  Those groups can then use the
  * notification event in whatever means they feel necessary.
-- 
2.7.4


  parent reply	other threads:[~2016-12-20 15:20 UTC|newest]

Thread overview: 4+ messages / expand[flat|nested]  mbox.gz  Atom feed  top
2016-12-20 15:20 [RFC][PATCH 0/2] fsnotify: super block watch Amir Goldstein
2016-12-20 15:20 ` [RFC][PATCH 1/2] fsnotify: add event mask FS_EVENT_ON_SB Amir Goldstein
2016-12-20 15:20 ` Amir Goldstein [this message]
2016-12-21 10:15 ` [RFC][PATCH 0/2] fsnotify: super block watch Marko Rauhamaa

Reply instructions:

You may reply publicly to this message via plain-text email
using any one of the following methods:

* Save the following mbox file, import it into your mail client,
  and reply-to-all from there: mbox

  Avoid top-posting and favor interleaved quoting:
  https://en.wikipedia.org/wiki/Posting_style#Interleaved_style

* Reply using the --to, --cc, and --in-reply-to
  switches of git-send-email(1):

  git send-email \
    --in-reply-to=1482247207-4424-3-git-send-email-amir73il@gmail.com \
    --to=amir73il@gmail.com \
    --cc=eparis@redhat.com \
    --cc=jack@suse.cz \
    --cc=linux-fsdevel@vger.kernel.org \
    --cc=linux-kernel@vger.kernel.org \
    --cc=viro@zeniv.linux.org.uk \
    /path/to/YOUR_REPLY

  https://kernel.org/pub/software/scm/git/docs/git-send-email.html

* If your mail client supports setting the In-Reply-To header
  via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line before the message body.
This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox;
as well as URLs for NNTP newsgroup(s).